import numpy as np
import pandas as pd
import xgboost as xgb
import os
from sklearn.mixture import GaussianMixture
import savic


from savic import SAVIC_Input_Sort
from savic import SAVIC_Core
from savic import SAVIC_CoreBeam
from savic import SAVIC_CoreAlpha
from savic import SAVIC_CoreBeamAlpha

def SAVIC(df_load_):
    
    df_s_ = []
    
    df_c_, df_cb_, df_ca_, df_cba_ = SAVIC_Input_Sort.SAVIC_Input_Sort(df_load_)
    
    if df_c_.size > 0:
        SAVIC_df_c_ = SAVIC_Core.SAVIC_Core(df_c_)
        df_s_.append(SAVIC_df_c_)
    if df_cb_.size > 0:
        SAVIC_df_cb_ = SAVIC_CoreBeam.SAVIC_CoreBeam(df_cb_)
        df_s_.append(SAVIC_df_cb_)
    if df_ca_.size > 0:
        SAVIC_df_ca_ = SAVIC_CoreAlpha.SAVIC_CoreAlpha(df_ca_)
        df_s_.append(SAVIC_df_ca_)
    if df_cba_.size > 0:
        SAVIC_df_cba_ = SAVIC_CoreBeamAlpha.SAVIC_CoreBeamAlpha(df_cba_)
        df_s_.append(SAVIC_df_cba_)
    SAVIC_df_ = pd.concat(df_s_)
    
    col_CBA_ = ['beta_par_core', 'alph_c', 'tau_b', 'alph_b', 'D_b', 'vv_b', 'tau_a', \
       'alph_a', 'D_a', 'vv_a', 'unstable', 'group', 'Pow_core', 'Pow_beam', \
       'Pow_alpha', 'kB_angle', 'ins_type']
    
    if df_cba_.size > 0:
        columns_ = col_CBA_
    elif (df_cb_.size > 0) & (df_ca_.size > 0):
        columns_ = col_CBA_
    elif df_cb_.size > 0:
        columns_ = SAVIC_df_cb_.columns
    elif df_ca_.size > 0:
        columns_ = SAVIC_df_ca_.columns
    else:
        columns_ = SAVIC_df_c_.columns
    
    SAVIC_df_ = SAVIC_df_[columns_]
    
    SAVIC_df_ = SAVIC_df_.reindex(df_load_.index)
    
    return SAVIC_df_
    
    


import numpy as np
import xgboost as xgb
import os
from sklearn.mixture import GaussianMixture
import savic


from savic import SAVIC_P_C
from savic import SAVIC_Q_C
from savic import SAVIC_C_C

def SAVIC_Core(df_c_):
    
    df_c_post_SP_ = SAVIC_P_C.SAVIC_P_C(df_c_)
    df_c_post_SQ_ = SAVIC_Q_C.SAVIC_Q_C(df_c_post_SP_)
    df_c_post_SC_ = SAVIC_C_C.SAVIC_C_C(df_c_post_SQ_)
    
    return df_c_post_SC_
    


import numpy as np
import xgboost as xgb
import os
from sklearn.mixture import GaussianMixture
import savic


from savic import SAVIC_P_CBA
from savic import SAVIC_Q_CBA
from savic import SAVIC_C_CBA

def SAVIC_CoreBeamAlpha(df_cba_):
    
    df_cba_post_SP_ = SAVIC_P_CBA.SAVIC_P_CBA(df_cba_)
    df_cba_post_SQ_ = SAVIC_Q_CBA.SAVIC_Q_CBA(df_cba_post_SP_)
    df_cba_post_SC_ = SAVIC_C_CBA.SAVIC_C_CBA(df_cba_post_SQ_)
    
    return df_cba_post_SC_
    


import numpy as np
import pandas as pd
import xgboost as xgb
import pandas as pd
import os
import savic

path_Q_CB = os.path.dirname(savic.__file__) + '/'

xgbc_kcb = xgb.XGBClassifier()
xgbc_kcb.load_model(path_Q_CB + 'Output/ML/models/xgbc_kcb.json')

xgbr_c1_b1 = xgb.XGBRegressor()
xgbr_c1_b1.load_model(path_Q_CB + 'Output/ML/models/xgbr_cb_c1_b1.json')

xgbr_c1_b0 = xgb.XGBRegressor()
xgbr_c1_b0.load_model(path_Q_CB + 'Output/ML/models/xgbr_cb_c1_b0.json')

xgbr_c0_b1_k0 = xgb.XGBRegressor()
xgbr_c0_b1_k0.load_model(path_Q_CB + 'Output/ML/models/xgbr_cb_c0_b1_k0.json')

xgbr_c0_b1_k1 = xgb.XGBRegressor()
xgbr_c0_b1_k1.load_model(path_Q_CB + 'Output/ML/models/xgbr_cb_c0_b1_k1.json')

def SAVIC_Q_CB(df_cb_):
    
    df_cb_uns_ = df_cb_[df_cb_['unstable']].drop(columns = ['unstable'])
    df_cb_uns_pre_SQ_ = df_cb_uns_.copy()
    df_cb_uns_pre_SQ_[['beta_par_core', 'alph_c', 'alph_b']] = \
    np.log10(df_cb_uns_pre_SQ_[['beta_par_core', 'alph_c', 'alph_b']])
    df_cb_uns_pre_SQ_.columns = [['log_beta_par_core', 'log_alph_c', 'tau_b', 'log_alph_b', 'D_b', 'vv_b']]
    
    df_cb_uns_postC_SQ_ = df_cb_uns_pre_SQ_.copy()
    df_cb_uns_postC_SQ_['group'] = xgbc_kcb.predict(df_cb_uns_pre_SQ_.values)
    
    df_cb_uns_preQ_SQ_ = df_cb_uns_postC_SQ_.copy()
    df_cb_uns_preQ_SQ_[['log_beta_par_core', 'log_alph_c', 'log_alph_b']] = \
    pow(10., df_cb_uns_preQ_SQ_[['log_beta_par_core', 'log_alph_c', 'log_alph_b']]) 
    df_cb_uns_preQ_SQ_.columns = ['beta_par_core', 'alph_c', 'tau_b', 'alph_b', 'D_b', 'vv_b', 'group']
    
    # regressors
    df_cb_uns_preQ_SQ_0_ = df_cb_uns_preQ_SQ_[df_cb_uns_preQ_SQ_['group'] == 0].drop(columns = ['group'])
    df_cb_uns_preQ_SQ_1_ = df_cb_uns_preQ_SQ_[df_cb_uns_preQ_SQ_['group'] == 1].drop(columns = ['group'])
    df_cb_uns_preQ_SQ_2_ = df_cb_uns_preQ_SQ_[df_cb_uns_preQ_SQ_['group'] == 2].drop(columns = ['group'])
    df_cb_uns_preQ_SQ_3_ = df_cb_uns_preQ_SQ_[df_cb_uns_preQ_SQ_['group'] == 3].drop(columns = ['group'])
    df_cb_uns_preQ_SQ_4_ = df_cb_uns_preQ_SQ_[df_cb_uns_preQ_SQ_['group'] == 4].drop(columns = ['group'])
    df_cb_uns_preQ_SQ_5_ = df_cb_uns_preQ_SQ_[df_cb_uns_preQ_SQ_['group'] == 5].drop(columns = ['group'])
    
    
    if df_cb_uns_preQ_SQ_0_.size > 0:
        df_cb_uns_preQ_SQ_0_log_ = df_cb_uns_preQ_SQ_0_.copy()

        df_cb_uns_preQ_SQ_0_log_[['beta_par_core', 'alph_c','alph_b']] = \
        np.log10(df_cb_uns_preQ_SQ_0_log_[['beta_par_core', 'alph_c','alph_b']])
        
        df_cb_uns_preQ_SQ_0_log_[['Pow_core', 'Pow_beam', 'kB_angle']] = \
        xgbr_c1_b1.predict(df_cb_uns_preQ_SQ_0_log_).astype(float)

        df_cb_uns_preQ_SQ_0_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_core', 'Pow_beam', 'kB_angle']] = \
        10**df_cb_uns_preQ_SQ_0_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_core', 'Pow_beam', 'kB_angle']]

        df_cb_uns_postQ_SQ_0_ = df_cb_uns_preQ_SQ_0_log_.copy()
        df_cb_uns_postQ_SQ_0_ = df_cb_uns_postQ_SQ_0_[['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle']]
    else:
        df_cb_uns_postQ_SQ_0_ = pd.DataFrame(columns = ['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle'])

    if df_cb_uns_preQ_SQ_1_.size > 0:
        df_cb_uns_preQ_SQ_1_log_ = df_cb_uns_preQ_SQ_1_.copy()


        df_cb_uns_preQ_SQ_1_log_[['beta_par_core', 'alph_c','alph_b']] = \
        np.log10(df_cb_uns_preQ_SQ_1_log_[['beta_par_core', 'alph_c','alph_b']])
        
        df_cb_uns_preQ_SQ_1_log_[['Pow_core', 'Pow_beam', 'kB_angle']] = \
        xgbr_c1_b1.predict(df_cb_uns_preQ_SQ_1_log_).astype(float)

        df_cb_uns_preQ_SQ_1_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_core', 'Pow_beam', 'kB_angle']] = \
        10**df_cb_uns_preQ_SQ_1_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_core', 'Pow_beam', 'kB_angle']]

        df_cb_uns_postQ_SQ_1_ = df_cb_uns_preQ_SQ_1_log_.copy()
        df_cb_uns_postQ_SQ_1_ = df_cb_uns_postQ_SQ_1_[['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle']]
    else:
        df_cb_uns_postQ_SQ_1_ = pd.DataFrame(columns = ['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle'])

    if df_cb_uns_preQ_SQ_2_.size > 0:
        df_cb_uns_preQ_SQ_2_log_ = df_cb_uns_preQ_SQ_2_.copy()


        df_cb_uns_preQ_SQ_2_log_[['beta_par_core', 'alph_c','alph_b']] = \
        np.log10(df_cb_uns_preQ_SQ_2_log_[['beta_par_core', 'alph_c','alph_b']])
        
        df_cb_uns_preQ_SQ_2_log_[['Pow_core', 'kB_angle']] = \
        xgbr_c1_b0.predict(df_cb_uns_preQ_SQ_2_log_).astype(float)

        df_cb_uns_preQ_SQ_2_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_core', 'kB_angle']] = \
        10**df_cb_uns_preQ_SQ_2_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_core', 'kB_angle']]

        df_cb_uns_postQ_SQ_2_ = df_cb_uns_preQ_SQ_2_log_.copy()
        df_cb_uns_postQ_SQ_2_['Pow_beam'] = 0
        df_cb_uns_postQ_SQ_2_ = df_cb_uns_postQ_SQ_2_[['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle']]
    else:
        df_cb_uns_postQ_SQ_2_ = pd.DataFrame(columns = ['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle'])

    if df_cb_uns_preQ_SQ_3_.size > 0:
        df_cb_uns_preQ_SQ_3_log_ = df_cb_uns_preQ_SQ_3_.copy()


        df_cb_uns_preQ_SQ_3_log_[['beta_par_core', 'alph_c','alph_b']] = \
        np.log10(df_cb_uns_preQ_SQ_3_log_[['beta_par_core', 'alph_c','alph_b']])
        
        df_cb_uns_preQ_SQ_3_log_[['Pow_core', 'kB_angle']] = \
        xgbr_c1_b0.predict(df_cb_uns_preQ_SQ_3_log_).astype(float)

        df_cb_uns_preQ_SQ_3_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_core', 'kB_angle']] = \
        10**df_cb_uns_preQ_SQ_3_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_core', 'kB_angle']]

        df_cb_uns_postQ_SQ_3_ = df_cb_uns_preQ_SQ_3_log_.copy()
        df_cb_uns_postQ_SQ_3_['Pow_beam'] = 0
        df_cb_uns_postQ_SQ_3_ = df_cb_uns_postQ_SQ_3_[['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle']]
    else:
        df_cb_uns_postQ_SQ_3_ = pd.DataFrame(columns = ['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle'])

    if df_cb_uns_preQ_SQ_4_.size > 0:
        df_cb_uns_preQ_SQ_4_log_ = df_cb_uns_preQ_SQ_4_.copy()


        df_cb_uns_preQ_SQ_4_log_[['beta_par_core', 'alph_c','alph_b']] = \
        np.log10(df_cb_uns_preQ_SQ_4_log_[['beta_par_core', 'alph_c','alph_b']])
        
        df_cb_uns_preQ_SQ_4_log_[['Pow_beam', 'kB_angle']] = \
        xgbr_c0_b1_k1.predict(df_cb_uns_preQ_SQ_4_log_).astype(float)

        df_cb_uns_preQ_SQ_4_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_beam', 'kB_angle']] = \
        10**df_cb_uns_preQ_SQ_4_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_beam', 'kB_angle']]

        df_cb_uns_postQ_SQ_4_ = df_cb_uns_preQ_SQ_4_log_.copy()
        df_cb_uns_postQ_SQ_4_['Pow_core'] = 0
        df_cb_uns_postQ_SQ_4_ = df_cb_uns_postQ_SQ_4_[['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle']]
    else:
        df_cb_uns_postQ_SQ_4_ = pd.DataFrame(columns = ['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle'])

    if df_cb_uns_preQ_SQ_5_.size > 0:
        df_cb_uns_preQ_SQ_5_log_ = df_cb_uns_preQ_SQ_5_.copy()


        df_cb_uns_preQ_SQ_5_log_[['beta_par_core', 'alph_c','alph_b']] = \
        np.log10(df_cb_uns_preQ_SQ_5_log_[['beta_par_core', 'alph_c','alph_b']])
        
        df_cb_uns_preQ_SQ_5_log_[['Pow_beam', 'kB_angle']] = \
        xgbr_c0_b1_k0.predict(df_cb_uns_preQ_SQ_5_log_).astype(float)

        df_cb_uns_preQ_SQ_5_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_beam', 'kB_angle']] = \
        10**df_cb_uns_preQ_SQ_5_log_[['beta_par_core', 'alph_c','alph_b', 'Pow_beam', 'kB_angle']]

        df_cb_uns_postQ_SQ_5_ = df_cb_uns_preQ_SQ_5_log_.copy()
        df_cb_uns_postQ_SQ_5_['Pow_core'] = 0
        df_cb_uns_postQ_SQ_5_ = df_cb_uns_postQ_SQ_5_[['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle']]
    else:
        df_cb_uns_postQ_SQ_5_ = pd.DataFrame(columns = ['beta_par_core', 'alph_c', 'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', 'kB_angle'])

    df_cb_uns_postQ_SQ_ = pd.concat([df_cb_uns_postQ_SQ_0_, df_cb_uns_postQ_SQ_1_, df_cb_uns_postQ_SQ_2_, \
                                 df_cb_uns_postQ_SQ_3_, df_cb_uns_postQ_SQ_4_, df_cb_uns_postQ_SQ_5_]).sort_index()
    
    df_cb_postQ_SQ_ = df_cb_.copy()

    df_cb_postQ_SQ_['group'] = np.nan
    df_cb_postQ_SQ_.loc[df_cb_uns_preQ_SQ_.index, 'group' ] = \
    df_cb_uns_preQ_SQ_['group']
    df_cb_postQ_SQ_['group'] = df_cb_postQ_SQ_['group']#.astype('Int64')

    df_cb_postQ_SQ_[['Pow_core', 'Pow_beam', 'kB_angle']] = np.nan
    df_cb_postQ_SQ_.loc[df_cb_uns_postQ_SQ_.index, ['Pow_core', 'Pow_beam', 'kB_angle'] ] = \
    df_cb_uns_postQ_SQ_[['Pow_core', 'Pow_beam', 'kB_angle'] ]
    
    return df_cb_postQ_SQ_


import numpy as np
import pandas as pd
import xgboost as xgb
import os
from sklearn.mixture import GaussianMixture
import savic

gmm_cb_name = os.path.dirname(savic.__file__) + '/Output/ML/models/GMM_CB/GMM_CB'

# reload
means = np.load(gmm_cb_name + '_means.npy')
covar = np.load(gmm_cb_name + '_covariances.npy')
gmm_cb = GaussianMixture(n_components = len(means), covariance_type='full')
gmm_cb.precisions_cholesky_ = np.linalg.cholesky(np.linalg.inv(covar))
gmm_cb.weights_ = np.load(gmm_cb_name + '_weights.npy')
gmm_cb.means_ = means
gmm_cb.covariances_ = covar


def SAVIC_C_CB(df_cb_):
    
    df_cb_uns_ = df_cb_[df_cb_['unstable']].drop(columns = ['unstable'])
    df_cb_uns_pre_SC_ = df_cb_uns_.copy()

    df_cb_uns_pre_SC_['uns_IC'] = (1. + 0.367 / ( df_cb_uns_pre_SC_['beta_par_core'] - 0.011 )**0.364) < df_cb_uns_pre_SC_['alph_c']
    df_cb_uns_pre_SC_['uns_Mirror'] = (1. + 0.702 / ( df_cb_uns_pre_SC_['beta_par_core'] + 0.009 )**0.674) < df_cb_uns_pre_SC_['alph_c']
    df_cb_uns_pre_SC_['uns_FM'] = (1. - 0.408 / ( df_cb_uns_pre_SC_['beta_par_core'] - 0.410 )**0.529) > df_cb_uns_pre_SC_['alph_c']
    df_cb_uns_pre_SC_['uns_OFH'] = (1. - 1.454 / ( df_cb_uns_pre_SC_['beta_par_core'] + 0.178 )**1.023) > df_cb_uns_pre_SC_['alph_c']
    df_cb_uns_pre_SC_['uns_core'] = df_cb_uns_pre_SC_['uns_OFH'] | df_cb_uns_pre_SC_['uns_FM'] | df_cb_uns_pre_SC_['uns_Mirror'] | df_cb_uns_pre_SC_['uns_IC']

    df_cb_uns_pre_SC_['beta_par_beam'] = df_cb_uns_pre_SC_['beta_par_core'] * df_cb_uns_pre_SC_['D_b'] / df_cb_uns_pre_SC_['tau_b']
    df_cb_uns_pre_SC_['lambda_FH_beta'] = ( (df_cb_uns_pre_SC_['beta_par_core'] + df_cb_uns_pre_SC_['beta_par_beam']) - \
    (df_cb_uns_pre_SC_['beta_par_core'] * df_cb_uns_pre_SC_['alph_c'] + df_cb_uns_pre_SC_['beta_par_beam'] * df_cb_uns_pre_SC_['alph_b']) ) / 2.
    df_cb_uns_pre_SC_['lambda_FH_beam'] = df_cb_uns_pre_SC_['D_b'] * df_cb_uns_pre_SC_['vv_b']**2.
    df_cb_uns_pre_SC_['lambda_FH'] = df_cb_uns_pre_SC_['lambda_FH_beta'] + df_cb_uns_pre_SC_['lambda_FH_beam']# + df_cb_uns_pre_SC_['lambda_FH_alpha']
    df_cb_uns_pre_SC_['lambda_FH'] = df_cb_uns_pre_SC_['lambda_FH'] > 1.

    df_cb_uns_pre_SC_['Pow_core'][df_cb_uns_pre_SC_['Pow_core'] < 0] = 0
    df_cb_uns_pre_SC_['Pow_beam'][df_cb_uns_pre_SC_['Pow_beam'] < 0] = 0

    
    column_list_cb_ = ['beta_par_core', 'alph_c', 'kB_angle', \
                       'tau_b', 'D_b', 'alph_b', 'vv_b', 'Pow_core', 'Pow_beam', \
                       'uns_IC', 'uns_FM', 'uns_OFH', 'uns_core', 'lambda_FH']

    df_cb_uns_pre_SC_cut_ = df_cb_uns_pre_SC_[column_list_cb_]
    
    df_cb_uns_pre_SC_cut_[['beta_par_core', 'alph_c']] = np.log10(df_cb_uns_pre_SC_cut_[['beta_par_core', 'alph_c']])
    df_cb_uns_pre_SC_cut_.rename(columns={"beta_par_core": "log_beta_par_core", "alph_c": "log_alph_c"}, inplace = True)
    
    df_cb_uns_post_SC_ = df_cb_uns_pre_SC_cut_.copy()
    df_cb_uns_post_SC_['cluster'] = gmm_cb.predict(df_cb_uns_pre_SC_cut_.values)
    
    df_cb_uns_post_SC_[['log_beta_par_core', 'log_alph_c']] = \
    pow(10., df_cb_uns_post_SC_[['log_beta_par_core', 'log_alph_c']])
    df_cb_uns_post_SC_.rename(columns={"log_beta_par_core": "beta_par_core", "log_alph_c": "alph_c"}, inplace = True)
    
    df_cb_uns_post_SC_['ins_type'] = np.nan
    ins_types_cb_ = ['Oblique FM (B); resonant with Core', 'IC (B); $T_\perp/T_{||} < 1$', 'Oblique FM (B)', \
                    'Oblique Firehose', 'IC (B); $T_\perp/T_{||} > 1$', 'Parallel Firehose', 'IC (B), unstable core', 'IC (C)']
    for cluster, ins_type in zip(np.arange(8), ins_types_cb_):
        df_cb_uns_post_SC_['ins_type'][df_cb_uns_post_SC_['cluster'] == cluster] = ins_type
    
    df_cb_post_SC_ = df_cb_.copy()
    df_cb_post_SC_['ins_type'] = np.nan
    df_cb_post_SC_.loc[df_cb_uns_post_SC_.index, ['ins_type']] = df_cb_uns_post_SC_['ins_type']
    
    return df_cb_post_SC_


import os
import numpy as np
import wget
import warnings
warnings.filterwarnings("ignore")

# ---------------------------
# Expected file sizes dictionary
# ---------------------------

file_sizes_models = {
    'GMM_C/GMM_C.png': 48100,
    'GMM_C/GMM_C_Brazil_4.png': 3594210,
    'GMM_C/GMM_C_covariances.npy': 1696,
    'GMM_C/GMM_C_means.npy': 352,
    'GMM_C/GMM_C_weights.npy': 160,
    'GMM_CA/GMM_CA.png': 64121,
    'GMM_CA/GMM_CA_Brazil_6.png': 1089962,
    'GMM_CA/GMM_CA_covariances.npy': 9536,
    'GMM_CA/GMM_CA_means.npy': 800,
    'GMM_CA/GMM_CA_weights.npy': 176,
    'GMM_CB/GMM_CB.png': 82851,
    'GMM_CB/GMM_CB_Brazil_8.png': 1196694,
    'GMM_CB/GMM_CB_covariances.npy': 12672,
    'GMM_CB/GMM_CB_means.npy': 1024,
    'GMM_CB/GMM_CB_weights.npy': 192,
    'GMM_CBA/GMM_CBA.png': 142428,
    'GMM_CBA/GMM_CBA_Brazil_14.png': 4463508,
    'GMM_CBA/GMM_CBA_Brazil_refined_12.png': 4009836,
    'GMM_CBA/GMM_CBA_covariances.npy': 40560,
    'GMM_CBA/GMM_CBA_means.npy': 2256,
    'GMM_CBA/GMM_CBA_refined.png': 113786,
    'GMM_CBA/GMM_CBA_weights.npy': 240,
    'xgbc_kca.json': 14293033,
    'xgbc_kca.png': 51725,
    'xgbc_kcb.json': 58960041,
    'xgbc_kcb.png': 76527,
    'xgbc_kcba.json': 16574099,
    'xgbc_kcba.png': 92406,
    'xgbc_sus_c.json': 1850667,
    'xgbc_sus_c.png': 40687,
    'xgbc_sus_ca.json': 11984789,
    'xgbc_sus_ca.png': 29898,
    'xgbc_sus_cb.json': 18950584,
    'xgbc_sus_cb.png': 38598,
    'xgbc_sus_cba.json': 30488776,
    'xgbc_sus_cba.png': 43568,
    'xgbr_c.json': 16387983,
    'xgbr_c.png': 91188,
    'xgbr_ca_c0_a1_k0.json': 45454962,
    'xgbr_ca_c0_a1_k0.png': 108944,
    'xgbr_ca_c1_a0_k0.json': 16626577,
    'xgbr_ca_c1_a0_k0.png': 92845,
    'xgbr_ca_c1_a1_k0.json': 12857830,
    'xgbr_ca_c1_a1_k0.png': 99268,
    'xgbr_ca_c1_a1_k1.json': 3712066,
    'xgbr_ca_c1_a1_k1.png': 118761,
    'xgbr_cba_c0_b0_a1.json': 75864632,
    'xgbr_cba_c0_b0_a1.png': 113565,
    'xgbr_cba_c0_b1_a0_k0.json': 26606376,
    'xgbr_cba_c0_b1_a0_k0.png': 92448,
    'xgbr_cba_c0_b1_a0_k1.json': 16586299,
    'xgbr_cba_c0_b1_a0_k1.png': 93464,
    'xgbr_cba_c0_b1_a1.json': 23243410,
    'xgbr_cba_c0_b1_a1.png': 106928,
    'xgbr_cba_c1_b0_a0.json': 50577428,
    'xgbr_cba_c1_b0_a0.png': 82779,
    'xgbr_cba_c1_b0_a1.json': 24188460,
    'xgbr_cba_c1_b0_a1.png': 106436,
    'xgbr_cba_c1_b1_a0.json': 62450387,
    'xgbr_cba_c1_b1_a0.png': 126609,
    'xgbr_cba_c1_b1_a1.json': 10690959,
    'xgbr_cba_c1_b1_a1.png': 115973,
    'xgbr_cb_c0_b1_k0.json': 49250105,
    'xgbr_cb_c0_b1_k0.png': 112469,
    'xgbr_cb_c0_b1_k1.json': 29677188,
    'xgbr_cb_c0_b1_k1.png': 94806,
    'xgbr_cb_c1_b0.json': 58917559,
    'xgbr_cb_c1_b0.png': 86685,
    'xgbr_cb_c1_b1.json': 72669964,
    'xgbr_cb_c1_b1.png': 143089,
}

# ---------------------------
# Download models
# ---------------------------

for name in MODEL_FILES:
    expected_size = file_sizes_models.get(name)
    download_if_missing(BASE_PATH_MODELS, BASE_URL_MODELS, name, expected_size)

# ---------------------------
# Download tutorial files
# ---------------------------

for name in TUTORIAL_FILES:
    download_if_missing(BASE_PATH_TUTORIAL, BASE_URL_TUTORIAL, name)



import numpy as np
import pandas as pd
import xgboost as xgb
import pandas as pd
import os
import savic

path_Q_CBA = os.path.dirname(savic.__file__) + '/'

xgbc_kcba = xgb.XGBClassifier()
xgbc_kcba.load_model(path_Q_CBA + 'Output/ML/models/xgbc_kcba.json')

xgbr_c1_b1_a1 = xgb.XGBRegressor()
xgbr_c1_b1_a1.load_model(path_Q_CBA + 'Output/ML/models/xgbr_cba_c1_b1_a1.json')

xgbr_c1_b1_a0 = xgb.XGBRegressor()
xgbr_c1_b1_a0.load_model(path_Q_CBA + 'Output/ML/models/xgbr_cba_c1_b1_a0.json')

xgbr_c1_b0_a1 = xgb.XGBRegressor()
xgbr_c1_b0_a1.load_model(path_Q_CBA + 'Output/ML/models/xgbr_cba_c1_b0_a1.json')

xgbr_c1_b0_a0 = xgb.XGBRegressor()
xgbr_c1_b0_a0.load_model(path_Q_CBA + 'Output/ML/models/xgbr_cba_c1_b0_a0.json')

xgbr_c0_b1_a1 = xgb.XGBRegressor()
xgbr_c0_b1_a1.load_model(path_Q_CBA + 'Output/ML/models/xgbr_cba_c0_b1_a1.json')

xgbr_c0_b1_a0_k1 = xgb.XGBRegressor()
xgbr_c0_b1_a0_k1.load_model(path_Q_CBA + 'Output/ML/models/xgbr_cba_c0_b1_a0_k1.json')

xgbr_c0_b1_a0_k0 = xgb.XGBRegressor()
xgbr_c0_b1_a0_k0.load_model(path_Q_CBA + 'Output/ML/models/xgbr_cba_c0_b1_a0_k0.json')

xgbr_c0_b0_a1 = xgb.XGBRegressor()
xgbr_c0_b0_a1.load_model(path_Q_CBA + 'Output/ML/models/xgbr_cba_c0_b0_a1.json')


def SAVIC_Q_CBA(df_cba_):
    
    df_cba_uns_ = df_cba_[df_cba_['unstable']].drop(columns = ['unstable'])
    
    df_cba_uns_pre_SQ_ = df_cba_uns_.copy()
    df_cba_uns_pre_SQ_[['beta_par_core', 'alph_c', 'alph_b', 'alph_a']] = \
    np.log10(df_cba_uns_pre_SQ_[['beta_par_core', 'alph_c', 'alph_b', 'alph_a']])
    df_cba_uns_pre_SQ_.columns = [['log_beta_par_core', 'log_alph_c', \
                                   'tau_b', 'log_alph_b', 'D_b', 'vv_b', \
                                   'tau_a', 'log_alph_a', 'D_a', 'vv_a']]
    
    df_cba_uns_postC_SQ_ = df_cba_uns_pre_SQ_.copy()
    df_cba_uns_postC_SQ_['group'] = xgbc_kcba.predict(df_cba_uns_pre_SQ_.values)
    
    df_cba_uns_preQ_SQ_ = df_cba_uns_postC_SQ_.copy()
    df_cba_uns_preQ_SQ_[['log_beta_par_core', 'log_alph_c', 'log_alph_b', 'log_alph_a']] = \
    pow(10., df_cba_uns_preQ_SQ_[['log_beta_par_core', 'log_alph_c', 'log_alph_b', 'log_alph_a']]) 
    df_cba_uns_preQ_SQ_.columns = ['beta_par_core', 'alph_c', \
                                   'tau_b', 'alph_b', 'D_b', 'vv_b', \
                                   'tau_a', 'alph_a', 'D_a', 'vv_a', 'group']

    
    df_cba_uns_preQ_SQ_0_ = df_cba_uns_preQ_SQ_[df_cba_uns_preQ_SQ_['group'] == 0].drop(columns = ['group'])
    df_cba_uns_preQ_SQ_1_ = df_cba_uns_preQ_SQ_[df_cba_uns_preQ_SQ_['group'] == 1].drop(columns = ['group'])
    df_cba_uns_preQ_SQ_2_ = df_cba_uns_preQ_SQ_[df_cba_uns_preQ_SQ_['group'] == 2].drop(columns = ['group'])
    df_cba_uns_preQ_SQ_3_ = df_cba_uns_preQ_SQ_[df_cba_uns_preQ_SQ_['group'] == 3].drop(columns = ['group'])
    df_cba_uns_preQ_SQ_4_ = df_cba_uns_preQ_SQ_[df_cba_uns_preQ_SQ_['group'] == 4].drop(columns = ['group'])
    df_cba_uns_preQ_SQ_5_ = df_cba_uns_preQ_SQ_[df_cba_uns_preQ_SQ_['group'] == 5].drop(columns = ['group'])
    df_cba_uns_preQ_SQ_6_ = df_cba_uns_preQ_SQ_[df_cba_uns_preQ_SQ_['group'] == 6].drop(columns = ['group'])
    df_cba_uns_preQ_SQ_7_ = df_cba_uns_preQ_SQ_[df_cba_uns_preQ_SQ_['group'] == 7].drop(columns = ['group'])
    
    
    # regressors
    cba_full_col_list_ = ['beta_par_core', 'alph_c', \
                          'tau_b', 'D_b', 'alph_b', 'vv_b', \
                          'tau_a', 'D_a', 'alph_a', 'vv_a', \
                          'Pow_core', 'Pow_beam', 'Pow_alpha', 'kB_angle']

    # group 0 - C1B1A1
    if df_cba_uns_preQ_SQ_0_.size > 0:
        df_cba_uns_preQ_SQ_0_log_ = df_cba_uns_preQ_SQ_0_.copy()

        df_cba_uns_preQ_SQ_0_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a']] = \
        np.log10(df_cba_uns_preQ_SQ_0_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a']])
        df_cba_uns_preQ_SQ_0_log_[['Pow_core', 'Pow_beam', 'Pow_alpha', 'kB_angle']] = \
        xgbr_c1_b1_a1.predict(df_cba_uns_preQ_SQ_0_log_).astype(float)

        df_cba_uns_preQ_SQ_0_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_core', 'Pow_beam', 'Pow_alpha', 'kB_angle']] = \
        10**df_cba_uns_preQ_SQ_0_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_core', 'Pow_beam', 'Pow_alpha', 'kB_angle']]

        df_cba_uns_postQ_SQ_0_ = df_cba_uns_preQ_SQ_0_log_.copy()
        df_cba_uns_postQ_SQ_0_ = df_cba_uns_postQ_SQ_0_[cba_full_col_list_]
    else:
        df_cba_uns_postQ_SQ_0_ = pd.DataFrame(columns = cba_full_col_list_)



    # group 1 - C1B1A0
    if df_cba_uns_preQ_SQ_1_.size > 0:
        df_cba_uns_preQ_SQ_1_log_ = df_cba_uns_preQ_SQ_1_.copy()

        df_cba_uns_preQ_SQ_1_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a']] = \
        np.log10(df_cba_uns_preQ_SQ_1_log_[['beta_par_core', 'alph_c', 'alph_b', 'alph_a']])

        df_cba_uns_preQ_SQ_1_log_[['Pow_core', 'Pow_beam', 'kB_angle']] = \
        xgbr_c1_b1_a0.predict(df_cba_uns_preQ_SQ_1_log_).astype(float)

        df_cba_uns_preQ_SQ_1_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_core', 'Pow_beam', 'kB_angle']] = \
        10**df_cba_uns_preQ_SQ_1_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_core', 'Pow_beam', 'kB_angle']]

        df_cba_uns_postQ_SQ_1_ = df_cba_uns_preQ_SQ_1_log_.copy()
        df_cba_uns_postQ_SQ_1_['Pow_alpha'] = 0
        df_cba_uns_postQ_SQ_1_ = df_cba_uns_postQ_SQ_1_[cba_full_col_list_]
    else:
        df_cba_uns_postQ_SQ_1_ = pd.DataFrame(columns = cba_full_col_list_)

    # group 2 - C1B0A1
    if df_cba_uns_preQ_SQ_2_.size > 0:
        df_cba_uns_preQ_SQ_2_log_ = df_cba_uns_preQ_SQ_2_.copy()

        df_cba_uns_preQ_SQ_2_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a']] = \
        np.log10(df_cba_uns_preQ_SQ_2_log_[['beta_par_core', 'alph_c', 'alph_b', 'alph_a']])

        df_cba_uns_preQ_SQ_2_log_[['Pow_core', 'Pow_alpha', 'kB_angle']] = \
        xgbr_c1_b0_a1.predict(df_cba_uns_preQ_SQ_2_log_).astype(float)

        df_cba_uns_preQ_SQ_2_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_core', 'Pow_alpha', 'kB_angle']] = \
        10**df_cba_uns_preQ_SQ_2_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_core', 'Pow_alpha', 'kB_angle']]

        df_cba_uns_postQ_SQ_2_ = df_cba_uns_preQ_SQ_2_log_.copy()
        df_cba_uns_postQ_SQ_2_['Pow_beam'] = 0
        df_cba_uns_postQ_SQ_2_ = df_cba_uns_postQ_SQ_2_[cba_full_col_list_]
    else:
        df_cba_uns_postQ_SQ_2_ = pd.DataFrame(columns = cba_full_col_list_)

    # group 3 - C1B0A0
    if df_cba_uns_preQ_SQ_3_.size > 0:
        df_cba_uns_preQ_SQ_3_log_ = df_cba_uns_preQ_SQ_3_.copy()

        df_cba_uns_preQ_SQ_3_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a']] = \
        np.log10(df_cba_uns_preQ_SQ_3_log_[['beta_par_core', 'alph_c', 'alph_b', 'alph_a']])

        df_cba_uns_preQ_SQ_3_log_[['Pow_core', 'kB_angle']] = \
        xgbr_c1_b0_a0.predict(df_cba_uns_preQ_SQ_3_log_).astype(float)

        df_cba_uns_preQ_SQ_3_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_core', 'kB_angle']] = \
        10**df_cba_uns_preQ_SQ_3_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_core', 'kB_angle']]

        df_cba_uns_postQ_SQ_3_ = df_cba_uns_preQ_SQ_3_log_.copy()
        df_cba_uns_postQ_SQ_3_[['Pow_beam', 'Pow_alpha']] = 0
        df_cba_uns_postQ_SQ_3_ = df_cba_uns_postQ_SQ_3_[cba_full_col_list_]
    else:
        df_cba_uns_postQ_SQ_3_ = pd.DataFrame(columns = cba_full_col_list_)

    # group 4 - C0B1A1
    if df_cba_uns_preQ_SQ_4_.size > 0:
        df_cba_uns_preQ_SQ_4_log_ = df_cba_uns_preQ_SQ_4_.copy()

        df_cba_uns_preQ_SQ_4_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a']] = \
        np.log10(df_cba_uns_preQ_SQ_4_log_[['beta_par_core', 'alph_c', 'alph_b', 'alph_a']])

        df_cba_uns_preQ_SQ_4_log_[['Pow_beam', 'Pow_alpha', 'kB_angle']] = \
        xgbr_c0_b1_a1.predict(df_cba_uns_preQ_SQ_4_log_).astype(float)

        df_cba_uns_preQ_SQ_4_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_beam', 'Pow_alpha', 'kB_angle']] = \
        10**df_cba_uns_preQ_SQ_4_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_beam', 'Pow_alpha', 'kB_angle']]

        df_cba_uns_postQ_SQ_4_ = df_cba_uns_preQ_SQ_4_log_.copy()
        df_cba_uns_postQ_SQ_4_[['Pow_core']] = 0
        df_cba_uns_postQ_SQ_4_ = df_cba_uns_postQ_SQ_4_[cba_full_col_list_]
    else:
        df_cba_uns_postQ_SQ_4_ = pd.DataFrame(columns = cba_full_col_list_)


    # group 5 - C0B1A0k1
    if df_cba_uns_preQ_SQ_5_.size > 0:
        df_cba_uns_preQ_SQ_5_log_ = df_cba_uns_preQ_SQ_5_.copy()

        df_cba_uns_preQ_SQ_5_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a']] = \
        np.log10(df_cba_uns_preQ_SQ_5_log_[['beta_par_core', 'alph_c', 'alph_b', 'alph_a']])

        df_cba_uns_preQ_SQ_5_log_[['Pow_beam', 'kB_angle']] = \
        xgbr_c0_b1_a0_k1.predict(df_cba_uns_preQ_SQ_5_log_).astype(float)

        df_cba_uns_preQ_SQ_5_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_beam', 'kB_angle']] = \
        10**df_cba_uns_preQ_SQ_5_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_beam', 'kB_angle']]

        df_cba_uns_postQ_SQ_5_ = df_cba_uns_preQ_SQ_5_log_.copy()
        df_cba_uns_postQ_SQ_5_[['Pow_core', 'Pow_alpha']] = 0
        df_cba_uns_postQ_SQ_5_ = df_cba_uns_postQ_SQ_5_[cba_full_col_list_]
    else:
        df_cba_uns_postQ_SQ_5_ = pd.DataFrame(columns = cba_full_col_list_)    

    # group 6 - C0B1A0k0
    if df_cba_uns_preQ_SQ_6_.size > 0:
        df_cba_uns_preQ_SQ_6_log_ = df_cba_uns_preQ_SQ_6_.copy()

        df_cba_uns_preQ_SQ_6_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a']] = \
        np.log10(df_cba_uns_preQ_SQ_6_log_[['beta_par_core', 'alph_c', 'alph_b', 'alph_a']])

        df_cba_uns_preQ_SQ_6_log_[['Pow_beam', 'kB_angle']] = \
        xgbr_c0_b1_a0_k0.predict(df_cba_uns_preQ_SQ_6_log_).astype(float)

        df_cba_uns_preQ_SQ_6_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_beam', 'kB_angle']] = \
        10**df_cba_uns_preQ_SQ_6_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_beam', 'kB_angle']]

        df_cba_uns_postQ_SQ_6_ = df_cba_uns_preQ_SQ_6_log_.copy()
        df_cba_uns_postQ_SQ_6_[['Pow_core', 'Pow_alpha']] = 0
        df_cba_uns_postQ_SQ_6_ = df_cba_uns_postQ_SQ_6_[cba_full_col_list_]
    else:
        df_cba_uns_postQ_SQ_6_ = pd.DataFrame(columns = cba_full_col_list_)    


    # group 7 - C0B0A1
    if df_cba_uns_preQ_SQ_7_.size > 0:
        df_cba_uns_preQ_SQ_7_log_ = df_cba_uns_preQ_SQ_7_.copy()

        df_cba_uns_preQ_SQ_7_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a']] = \
        np.log10(df_cba_uns_preQ_SQ_7_log_[['beta_par_core', 'alph_c', 'alph_b', 'alph_a']])

        df_cba_uns_preQ_SQ_7_log_[['Pow_alpha', 'kB_angle']] = \
        xgbr_c0_b0_a1.predict(df_cba_uns_preQ_SQ_7_log_).astype(float)

        df_cba_uns_preQ_SQ_7_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_alpha', 'kB_angle']] = \
        10**df_cba_uns_preQ_SQ_7_log_[['beta_par_core', 'alph_c', 'alph_b','alph_a', 'Pow_alpha', 'kB_angle']]

        df_cba_uns_postQ_SQ_7_ = df_cba_uns_preQ_SQ_7_log_.copy()
        df_cba_uns_postQ_SQ_7_[['Pow_core', 'Pow_beam']] = 0
        df_cba_uns_postQ_SQ_7_ = df_cba_uns_postQ_SQ_7_[cba_full_col_list_]
    else:
        df_cba_uns_postQ_SQ_7_ = pd.DataFrame(columns = cba_full_col_list_)    


    df_cba_uns_postQ_SQ_ = pd.concat([df_cba_uns_postQ_SQ_0_, df_cba_uns_postQ_SQ_1_, df_cba_uns_postQ_SQ_2_, \
                                      df_cba_uns_postQ_SQ_3_, df_cba_uns_postQ_SQ_4_, df_cba_uns_postQ_SQ_5_, \
                                      df_cba_uns_postQ_SQ_6_,df_cba_uns_postQ_SQ_7_]).sort_index()
    
    df_cba_postQ_SQ_ = df_cba_.copy()

    df_cba_postQ_SQ_['group'] = np.nan
    df_cba_postQ_SQ_.loc[df_cba_uns_preQ_SQ_.index, 'group' ] = \
    df_cba_uns_preQ_SQ_['group']
    df_cba_postQ_SQ_['group'] = df_cba_postQ_SQ_['group']#.astype('Int64')

    df_cba_postQ_SQ_[['Pow_core', 'Pow_beam', 'Pow_alpha', 'kB_angle']] = np.nan
    df_cba_postQ_SQ_.loc[df_cba_uns_postQ_SQ_.index, ['Pow_core', 'Pow_beam', 'Pow_alpha', 'kB_angle'] ] = \
    df_cba_uns_postQ_SQ_[['Pow_core', 'Pow_beam', 'Pow_alpha', 'kB_angle'] ]
    
    return df_cba_postQ_SQ_
    


import numpy as np
import pandas as pd
import xgboost as xgb
import os
from sklearn.mixture import GaussianMixture
import savic

gmm_c_name = os.path.dirname(savic.__file__) + '/Output/ML/models/GMM_C/GMM_C'

# reload
means = np.load(gmm_c_name + '_means.npy')
covar = np.load(gmm_c_name + '_covariances.npy')
gmm_c = GaussianMixture(n_components = len(means), covariance_type='full')
gmm_c.precisions_cholesky_ = np.linalg.cholesky(np.linalg.inv(covar))
gmm_c.weights_ = np.load(gmm_c_name + '_weights.npy')
gmm_c.means_ = means
gmm_c.covariances_ = covar

def SAVIC_C_C(df_c_):
    
    df_c_uns_ = df_c_[df_c_['unstable']].drop(columns = ['unstable'])
    df_c_uns_.drop(columns = ['Pow_core'], inplace = True)
    
    df_c_uns_pre_SC_ = df_c_uns_.copy()
    df_c_uns_pre_SC_['uns_IC'] = (1. + 0.367 / ( df_c_uns_pre_SC_['beta_par_core'] - 0.011 )**0.364) < df_c_uns_pre_SC_['alph_c']
    df_c_uns_pre_SC_['uns_Mirror'] = (1. + 0.702 / ( df_c_uns_pre_SC_['beta_par_core'] + 0.009 )**0.674) < df_c_uns_pre_SC_['alph_c']
    df_c_uns_pre_SC_['uns_FM'] = (1. - 0.408 / ( df_c_uns_pre_SC_['beta_par_core'] - 0.410 )**0.529) > df_c_uns_pre_SC_['alph_c']
    df_c_uns_pre_SC_['uns_OFH'] = (1. - 1.454 / ( df_c_uns_pre_SC_['beta_par_core'] + 0.178 )**1.023) > df_c_uns_pre_SC_['alph_c']
    df_c_uns_pre_SC_['uns_core'] = df_c_uns_pre_SC_['uns_OFH'] | df_c_uns_pre_SC_['uns_FM'] | df_c_uns_pre_SC_['uns_Mirror'] | df_c_uns_pre_SC_['uns_IC']
    df_c_uns_pre_SC_.drop(columns = ['uns_Mirror'], inplace = True)
    
    df_c_uns_pre_SC_[['beta_par_core', 'alph_c']] = np.log10(df_c_uns_pre_SC_[['beta_par_core', 'alph_c']])
    df_c_uns_pre_SC_.rename(columns={"beta_par_core": "log_beta_par_core", "alph_c": "log_alph_c"}, inplace = True)
    
    df_c_uns_post_SC_ = df_c_uns_pre_SC_.copy()
    df_c_uns_post_SC_['cluster'] = gmm_c.predict(df_c_uns_pre_SC_[['log_beta_par_core', 'log_alph_c', 'kB_angle', \
                                                                   'uns_IC', 'uns_FM', 'uns_OFH', 'uns_core']].values)
    
    df_c_uns_post_SC_[['log_beta_par_core', 'log_alph_c']] = \
    pow(10., df_c_uns_post_SC_[['log_beta_par_core', 'log_alph_c']])
    df_c_uns_post_SC_.rename(columns={"log_beta_par_core": "beta_par_core", "log_alph_c": "alph_c"}, inplace = True)
    
    df_c_uns_post_SC_['ins_type'] = np.nan
    ins_types = ['Ion Cyclotron', 'Parallel Firehose', 'Mirror', 'Oblique Firehose']
    for cluster, ins_type in zip(np.arange(4), ins_types):
        df_c_uns_post_SC_['ins_type'][df_c_uns_post_SC_['cluster'] == cluster] = ins_type
    
    df_c_post_SC_ = df_c_.copy()
    df_c_post_SC_['ins_type'] = np.nan
    df_c_post_SC_.loc[df_c_uns_post_SC_.index, ['ins_type']] = df_c_uns_post_SC_['ins_type']
    
    return df_c_post_SC_


import numpy as np
import pandas as pd
import xgboost as xgb
import os
from sklearn.mixture import GaussianMixture
import savic

gmm_ca_name = os.path.dirname(savic.__file__) + '/Output/ML/models/GMM_CA/GMM_CA'

# reload
means = np.load(gmm_ca_name + '_means.npy')
covar = np.load(gmm_ca_name + '_covariances.npy')
gmm_ca = GaussianMixture(n_components = len(means), covariance_type='full')
gmm_ca.precisions_cholesky_ = np.linalg.cholesky(np.linalg.inv(covar))
gmm_ca.weights_ = np.load(gmm_ca_name + '_weights.npy')
gmm_ca.means_ = means
gmm_ca.covariances_ = covar


def SAVIC_C_CA(df_ca_):
    
    df_ca_uns_ = df_ca_[df_ca_['unstable']].drop(columns = ['unstable'])
    
    df_ca_uns_pre_SC_ = df_ca_uns_.copy()

    df_ca_uns_pre_SC_['uns_IC'] = (1. + 0.367 / ( df_ca_uns_pre_SC_['beta_par_core'] - 0.011 )**0.364) < df_ca_uns_pre_SC_['alph_c']
    df_ca_uns_pre_SC_['uns_Mirror'] = (1. + 0.702 / ( df_ca_uns_pre_SC_['beta_par_core'] + 0.009 )**0.674) < df_ca_uns_pre_SC_['alph_c']
    df_ca_uns_pre_SC_['uns_FM'] = (1. - 0.408 / ( df_ca_uns_pre_SC_['beta_par_core'] - 0.410 )**0.529) > df_ca_uns_pre_SC_['alph_c']
    df_ca_uns_pre_SC_['uns_OFH'] = (1. - 1.454 / ( df_ca_uns_pre_SC_['beta_par_core'] + 0.178 )**1.023) > df_ca_uns_pre_SC_['alph_c']
    df_ca_uns_pre_SC_['uns_core'] = df_ca_uns_pre_SC_['uns_OFH'] | df_ca_uns_pre_SC_['uns_FM'] | df_ca_uns_pre_SC_['uns_Mirror'] | df_ca_uns_pre_SC_['uns_IC']

    df_ca_uns_pre_SC_['beta_par_alpha'] = df_ca_uns_pre_SC_['beta_par_core'] * df_ca_uns_pre_SC_['D_a'] / df_ca_uns_pre_SC_['tau_a']
    df_ca_uns_pre_SC_['lambda_FH_beta'] = ( (df_ca_uns_pre_SC_['beta_par_core'] + df_ca_uns_pre_SC_['beta_par_alpha']) - \
    (df_ca_uns_pre_SC_['beta_par_core'] * df_ca_uns_pre_SC_['alph_c'] + df_ca_uns_pre_SC_['beta_par_alpha'] * df_ca_uns_pre_SC_['alph_a']) ) / 2.
    df_ca_uns_pre_SC_['lambda_FH_alpha'] = 4 * df_ca_uns_pre_SC_['D_a'] / (1. + df_ca_uns_pre_SC_['D_a'] + 4 * df_ca_uns_pre_SC_['D_a']) * df_ca_uns_pre_SC_['vv_a']**2.
    df_ca_uns_pre_SC_['lambda_FH'] = df_ca_uns_pre_SC_['lambda_FH_beta'] + df_ca_uns_pre_SC_['lambda_FH_alpha']
    df_ca_uns_pre_SC_['lambda_FH'] = df_ca_uns_pre_SC_['lambda_FH'] > 1.

    df_ca_uns_pre_SC_['Pow_core'][df_ca_uns_pre_SC_['Pow_core'] < 0] = 0
    df_ca_uns_pre_SC_['Pow_alpha'][df_ca_uns_pre_SC_['Pow_alpha'] < 0] = 0
    
    column_list_ca_ = ['beta_par_core', 'alph_c', 'kB_angle', \
                       'tau_a', 'D_a', 'alph_a', 'vv_a', 'Pow_core', 'Pow_alpha', \
                       'uns_IC', 'uns_FM', 'uns_OFH', 'uns_core', 'lambda_FH']

    df_ca_uns_pre_SC_cut_ = df_ca_uns_pre_SC_[column_list_ca_]
    
    df_ca_uns_pre_SC_cut_[['beta_par_core', 'alph_c']] = np.log10(df_ca_uns_pre_SC_cut_[['beta_par_core', 'alph_c']])
    df_ca_uns_pre_SC_cut_.rename(columns={"beta_par_core": "log_beta_par_core", "alph_c": "log_alph_c"}, inplace = True)
    
    df_ca_uns_post_SC_ = df_ca_uns_pre_SC_cut_.copy()
    df_ca_uns_post_SC_['cluster'] = gmm_ca.predict(df_ca_uns_pre_SC_cut_.values)
    
    df_ca_uns_post_SC_['ins_type'] = np.nan
    ins_types_ca_ = ['Parallel Firehose', 'IC (A)', 'Oblique Firehose', \
                     'CGL Firehose; Mirror', 'A anis; borderline PFH', 'IC (C)']
    for cluster, ins_type in zip(np.arange(6), ins_types_ca_):
        df_ca_uns_post_SC_['ins_type'][df_ca_uns_post_SC_['cluster'] == cluster] = ins_type
    
    df_ca_post_SC_ = df_ca_.copy()
    df_ca_post_SC_['ins_type'] = np.nan
    df_ca_post_SC_.loc[df_ca_uns_post_SC_.index, ['ins_type']] = df_ca_uns_post_SC_['ins_type']
    
    return df_ca_post_SC_


import savic
import os

def tutorial_path():
    
    return os.path.dirname(savic.__file__) + '/tutorial/SAVIC_Examples.h5'
    
def readme_path():
    
    return os.path.dirname(savic.__file__) + '/tutorial/SAVIC_readme.pdf'
    
def test_path():
    
    return os.path.dirname(savic.__file__) + '/tutorial/SAVIC_testing.ipynb'

def article_path():
    
    return [os.path.dirname(savic.__file__) + '/tutorial/Article_I_Statistical_Trends.pdf', os.path.dirname(savic.__file__) + '/tutorial/Article_II_Classification_and_Multidimensional_Mapping.pdf']
    

def docs_path():
    
    return 'https://savic.readthedocs.io/en/latest/'





import numpy as np
import xgboost as xgb
import os
from sklearn.mixture import GaussianMixture
import savic


from savic import SAVIC_P_CA
from savic import SAVIC_Q_CA
from savic import SAVIC_C_CA

def SAVIC_CoreAlpha(df_ca_):
    
    df_ca_post_SP_ = SAVIC_P_CA.SAVIC_P_CA(df_ca_)
    df_ca_post_SQ_ = SAVIC_Q_CA.SAVIC_Q_CA(df_ca_post_SP_)
    df_ca_post_SC_ = SAVIC_C_CA.SAVIC_C_CA(df_ca_post_SQ_)
    
    return df_ca_post_SC_
    


import numpy as np
import pandas as pd
import xgboost as xgb
import os
import savic

path_P_CB = os.path.dirname(savic.__file__) + '/Output/ML/models/xgbc_sus_cb.json'
xgbc_sus_cb = xgb.XGBClassifier()
xgbc_sus_cb.load_model(path_P_CB)

def SAVIC_P_CB(df_cb_):
    df_cb_pre_SP_ = df_cb_.copy()
    df_cb_pre_SP_[['beta_par_core', 'alph_c', 'tau_b', 'alph_b']] = \
    np.log10(df_cb_pre_SP_[['beta_par_core', 'alph_c', 'tau_b', 'alph_b']])
    df_cb_pre_SP_['unstable'] = xgbc_sus_cb.predict(df_cb_pre_SP_.values).astype('bool')
    
    df_cb_post_SP_ = df_cb_pre_SP_.copy()
    df_cb_post_SP_[['beta_par_core', 'alph_c', 'tau_b', 'alph_b']] = \
    pow(10., df_cb_post_SP_[['beta_par_core', 'alph_c', 'tau_b', 'alph_b']])
    
    return df_cb_post_SP_


import numpy as np
import pandas as pd
import xgboost as xgb
import os
import savic

path_P_C = os.path.dirname(savic.__file__) + '/Output/ML/models/xgbc_sus_c.json'
xgbc_sus_c = xgb.XGBClassifier()
xgbc_sus_c.load_model(path_P_C)

def SAVIC_P_C(df_c_):
    
    df_c_pre_SP_ = df_c_.copy()
    df_c_pre_SP_ = np.log10(df_c_pre_SP_)
    df_c_pre_SP_['unstable'] = xgbc_sus_c.predict(df_c_pre_SP_.values).astype('bool')
    
    df_c_post_SP_ = df_c_pre_SP_.copy()
    df_c_post_SP_[['beta_par_core', 'alph_c']] = pow(10., df_c_post_SP_[['beta_par_core', 'alph_c']])
    
    return df_c_post_SP_


import numpy as np
import pandas as pd
import xgboost as xgb
import os
import savic

path_P_CA = os.path.dirname(savic.__file__) + '/Output/ML/models/xgbc_sus_ca.json'
xgbc_sus_ca = xgb.XGBClassifier()
xgbc_sus_ca.load_model(path_P_CA)

def SAVIC_P_CA(df_ca_):
    df_ca_pre_SP_ = df_ca_.copy()
    df_ca_pre_SP_[['beta_par_core', 'alph_c', 'tau_a', 'alph_a']] = \
    np.log10(df_ca_pre_SP_[['beta_par_core', 'alph_c', 'tau_a', 'alph_a']])
    df_ca_pre_SP_['unstable'] = xgbc_sus_ca.predict(df_ca_pre_SP_.values).astype('bool')
    
    df_ca_post_SP_ = df_ca_pre_SP_.copy()
    df_ca_post_SP_[['beta_par_core', 'alph_c', 'tau_a', 'alph_a']] = \
    pow(10., df_ca_post_SP_[['beta_par_core', 'alph_c', 'tau_a', 'alph_a']])
    
    return df_ca_post_SP_


import numpy as np
import pandas as pd
import xgboost as xgb
import os
import savic

path_P_CBA = os.path.dirname(savic.__file__) + '/Output/ML/models/xgbc_sus_cba.json'
xgbc_sus_cba = xgb.XGBClassifier()
xgbc_sus_cba.load_model(path_P_CBA)

def SAVIC_P_CBA(df_cba_):
    df_cba_pre_SP_ = df_cba_.copy()
    df_cba_pre_SP_['unstable'] = xgbc_sus_cba.predict(df_cba_pre_SP_.values).astype('bool')
    df_cba_post_SP_ = df_cba_pre_SP_.copy()
    
    return df_cba_post_SP_
    


import numpy as np
import pandas as pd
import xgboost as xgb
import os
import savic

path_Q_C = os.path.dirname(savic.__file__) + '/Output/ML/models/xgbr_c.json'
xgbr_c1_c = xgb.XGBRegressor()
xgbr_c1_c.load_model(path_Q_C)

def SAVIC_Q_C(df_c_):
    
    df_c_uns_ = df_c_[df_c_['unstable']].drop(columns = ['unstable'])
    
    df_c_uns_pre_SQ_ = df_c_uns_.copy()
    df_c_uns_pre_SQ_ = np.log10(df_c_uns_pre_SQ_)
    df_c_uns_pre_SQ_.columns = [['log_beta_par_core', 'log_alph_c']]
    
    df_c_uns_pre_SQ_[['log_Pow_core', 'log_kB_angle']] = xgbr_c1_c.predict(df_c_uns_pre_SQ_.values)
    
    df_c_uns_post_SQ_ = pow(10., df_c_uns_pre_SQ_)
    df_c_uns_post_SQ_.columns = ['beta_par_core', 'alph_c', 'Pow_core', 'kB_angle']
    
    df_c_post_SQ_ = df_c_.copy()
    df_c_post_SQ_[['Pow_core', 'kB_angle']] = np.nan
    df_c_post_SQ_.loc[df_c_uns_post_SQ_.index, ['Pow_core', 'kB_angle']] = df_c_uns_post_SQ_[['Pow_core', 'kB_angle']]
    
    return df_c_post_SQ_


import numpy as np
import xgboost as xgb
import os
from sklearn.mixture import GaussianMixture
import savic


from savic import SAVIC_P_CB
from savic import SAVIC_Q_CB
from savic import SAVIC_C_CB

def SAVIC_CoreBeam(df_cb_):
    
    df_cb_post_SP_ = SAVIC_P_CB.SAVIC_P_CB(df_cb_)
    df_cb_post_SQ_ = SAVIC_Q_CB.SAVIC_Q_CB(df_cb_post_SP_)
    df_cb_post_SC_ = SAVIC_C_CB.SAVIC_C_CB(df_cb_post_SQ_)
    
    return df_cb_post_SC_
    


import numpy as np
import pandas as pd
import xgboost as xgb
import os
import savic

def SAVIC_Input_Sort(df_load_):
    
    cols_c_ = ['beta_par_core', 'alph_c']
    cols_cb_ = ['beta_par_core', 'alph_c', 'tau_b', 'alph_b', 'D_b', 'vv_b']
    cols_ca_ = ['beta_par_core', 'alph_c', 'tau_a', 'alph_a', 'D_a', 'vv_a']
    cols_cba_ = ['beta_par_core', 'alph_c', 'tau_b', 'alph_b', 'D_b', 'vv_b', 'tau_a', 'alph_a', 'D_a', 'vv_a']
    
    df_c_empty = pd.DataFrame(columns = cols_c_)
    df_cb_empty = pd.DataFrame(columns = cols_cb_)
    df_ca_empty = pd.DataFrame(columns = cols_ca_)
    df_cba_empty = pd.DataFrame(columns = cols_cba_)
    
    if set(df_load_.columns) == set(cols_cba_):
        df_c_ = df_load_[~(df_load_['D_b'] > 0) & ~(df_load_['D_a'] > 0)][cols_c_]
        df_cb_ = df_load_[(df_load_['D_b'] > 0) & ~(df_load_['D_a'] > 0)][cols_cb_]
        df_ca_ = df_load_[~(df_load_['D_b'] > 0) & (df_load_['D_a'] > 0)][cols_ca_]
        df_cba_ = df_load_[(df_load_['D_b'] > 0) & (df_load_['D_a'] > 0)]
        if df_c_.size == 0:
            df_c_ = df_c_empty
        if df_cb_.size == 0:
            df_cb_ = df_cb_empty
        if df_ca_.size == 0:
            df_ca_ = df_ca_empty
        if df_cba_.size == 0:
            df_cba_ = df_cba_empty
        return df_c_, df_cb_, df_ca_, df_cba_
    elif set(df_load_.columns) == set(cols_ca_):
        df_c_ = df_load_[~(df_load_['D_a'] > 0)][cols_c_]
        df_ca_ = df_load_[df_load_['D_a'] > 0]
        if df_c_.size == 0:
            df_c_ = df_c_empty
        if df_ca_.size == 0:
            df_ca_ = df_ca_empty
        df_cb_ = df_cb_empty
        df_cba_ = df_cba_empty
        return df_c_, df_cb_, df_ca_, df_cba_
    elif set(df_load_.columns) == set(cols_cb_):
        df_c_ = df_load_[~(df_load_['D_b'] > 0)][cols_c_]
        df_cb_ = df_load_[df_load_['D_b'] > 0]
        if df_c_.size == 0:
            df_c_ = df_c_empty
        if df_cb_.size == 0:
            df_cb_ = df_cb_empty
        df_ca_ = df_ca_empty
        df_cba_ = df_cba_empty
        return df_c_, df_cb_, df_ca_, df_cba_
    elif set(df_load_.columns) == set(cols_c_):
        df_c_ = df_load_
        if df_c_.size == 0:
            df_c_ = df_c_empty
        df_cb_ = df_cb_empty
        df_ca_ = df_ca_empty
        df_cba_ = df_cba_empty
        return df_c_, df_cb_, df_ca_, df_cba_
    else:
        raise ValueError('Column set not recognized')
        
        


import numpy as np
import pandas as pd
import xgboost as xgb
import os
from sklearn.mixture import GaussianMixture
import savic

gmm_cba_name = os.path.dirname(savic.__file__) + '/Output/ML/models/GMM_CBA/GMM_CBA'

# reload
means = np.load(gmm_cba_name + '_means.npy')
covar = np.load(gmm_cba_name + '_covariances.npy')
gmm_cba = GaussianMixture(n_components = len(means), covariance_type='full')
gmm_cba.precisions_cholesky_ = np.linalg.cholesky(np.linalg.inv(covar))
gmm_cba.weights_ = np.load(gmm_cba_name + '_weights.npy')
gmm_cba.means_ = means
gmm_cba.covariances_ = covar


def SAVIC_C_CBA(df_cba_):
    df_cba_uns_ = df_cba_[df_cba_['unstable']].drop(columns = ['unstable'])
    
    df_cba_uns_pre_SC_ = df_cba_uns_.copy()

    df_cba_uns_pre_SC_['uns_IC'] = (1. + 0.367 / ( df_cba_uns_pre_SC_['beta_par_core'] - 0.011 )**0.364) < df_cba_uns_pre_SC_['alph_c']
    df_cba_uns_pre_SC_['uns_Mirror'] = (1. + 0.702 / ( df_cba_uns_pre_SC_['beta_par_core'] + 0.009 )**0.674) < df_cba_uns_pre_SC_['alph_c']
    df_cba_uns_pre_SC_['uns_FM'] = (1. - 0.408 / ( df_cba_uns_pre_SC_['beta_par_core'] - 0.410 )**0.529) > df_cba_uns_pre_SC_['alph_c']
    df_cba_uns_pre_SC_['uns_OFH'] = (1. - 1.454 / ( df_cba_uns_pre_SC_['beta_par_core'] + 0.178 )**1.023) > df_cba_uns_pre_SC_['alph_c']
    df_cba_uns_pre_SC_['uns_core'] = df_cba_uns_pre_SC_['uns_OFH'] | df_cba_uns_pre_SC_['uns_FM'] | df_cba_uns_pre_SC_['uns_Mirror'] | df_cba_uns_pre_SC_['uns_IC']


    df_cba_uns_pre_SC_['beta_par_beam'] = df_cba_uns_pre_SC_['beta_par_core'] * df_cba_uns_pre_SC_['D_b'] / df_cba_uns_pre_SC_['tau_b']
    df_cba_uns_pre_SC_['beta_par_alpha'] = df_cba_uns_pre_SC_['beta_par_core'] * df_cba_uns_pre_SC_['D_a'] / df_cba_uns_pre_SC_['tau_a']

    df_cba_uns_pre_SC_['lambda_FH_beta'] = ( (df_cba_uns_pre_SC_['beta_par_core'] + df_cba_uns_pre_SC_['beta_par_beam'] + df_cba_uns_pre_SC_['beta_par_alpha']) - \
    (df_cba_uns_pre_SC_['beta_par_core'] * df_cba_uns_pre_SC_['alph_c'] + df_cba_uns_pre_SC_['beta_par_beam'] * df_cba_uns_pre_SC_['alph_b'] + \
    df_cba_uns_pre_SC_['beta_par_alpha'] * df_cba_uns_pre_SC_['alph_a']) ) / 2.

    df_cba_uns_pre_SC_['lambda_FH_alpha'] = 4 * df_cba_uns_pre_SC_['D_a'] / (1. + df_cba_uns_pre_SC_['D_b'] + 4 * df_cba_uns_pre_SC_['D_a']) * df_cba_uns_pre_SC_['vv_a']**2.
    df_cba_uns_pre_SC_['lambda_FH_beam'] = df_cba_uns_pre_SC_['D_b'] / (1. + df_cba_uns_pre_SC_['D_b'] + 4 * df_cba_uns_pre_SC_['D_a']) * df_cba_uns_pre_SC_['vv_b']**2.
    df_cba_uns_pre_SC_['lambda_FH'] = df_cba_uns_pre_SC_['lambda_FH_beta'] + df_cba_uns_pre_SC_['lambda_FH_alpha'] + df_cba_uns_pre_SC_['lambda_FH_beam']
    df_cba_uns_pre_SC_['lambda_FH'] = df_cba_uns_pre_SC_['lambda_FH'] > 1.

    df_cba_uns_pre_SC_['Pow_core'][df_cba_uns_pre_SC_['Pow_core'] < 0] = 0
    df_cba_uns_pre_SC_['Pow_beam'][df_cba_uns_pre_SC_['Pow_beam'] < 0] = 0
    df_cba_uns_pre_SC_['Pow_alpha'][df_cba_uns_pre_SC_['Pow_alpha'] < 0] = 0

    column_list_cba_ = ['beta_par_core', 'alph_c', 'kB_angle', \
                        'tau_b', 'alph_b', 'D_b', 'vv_b', \
                        'tau_a', 'D_a', 'alph_a', 'vv_a', 'Pow_core', 'Pow_beam', 'Pow_alpha', \
                        'uns_IC', 'uns_FM', 'uns_OFH', 'uns_core', 'lambda_FH']

    df_cba_uns_pre_SC_cut_ = df_cba_uns_pre_SC_[column_list_cba_]
    
    df_cba_uns_pre_SC_cut_[['beta_par_core', 'alph_c']] = np.log10(df_cba_uns_pre_SC_cut_[['beta_par_core', 'alph_c']])
    df_cba_uns_pre_SC_cut_.rename(columns={"beta_par_core": "log_beta_par_core", "alph_c": "log_alph_c"}, inplace = True)
    
    df_cba_uns_post_SC_ = df_cba_uns_pre_SC_cut_.copy()
    df_cba_uns_post_SC_['cluster'] = gmm_cba.predict(df_cba_uns_pre_SC_cut_.values)
    
    df_cba_uns_post_SC_[['log_beta_par_core', 'log_alph_c']] = \
    pow(10., df_cba_uns_post_SC_[['log_beta_par_core', 'log_alph_c']])
    df_cba_uns_post_SC_.rename(columns={"log_beta_par_core": "beta_par_core", "log_alph_c": "alph_c"}, inplace = True)
    
    df_cba_uns_post_SC_['r_cluster'] = np.nan
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 0] = 9
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 1] = 2
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 2] = 1
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 3] = 4
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 4] = 7
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 5] = 5
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 6] = 12
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 7] = 1
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 8] = 11
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 9] = 10
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 10] = 7
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 11] = 6
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 12] = 8
    df_cba_uns_post_SC_['r_cluster'][df_cba_uns_post_SC_['cluster'] == 13] = 3

    df_cba_uns_post_SC_['r_cluster'] = df_cba_uns_post_SC_['r_cluster'].astype(int)
    
    df_cba_uns_post_SC_['ins_type'] = np.nan
    ins_types_cba_ = ['IC (C)', 'IC (C); A unstable', \
                      'IC (B), C unstable', 'IC (B); A unstable', 'IC (B); high B anis', 'IC (B); borderline PFH', \
                      'IC (A)', 'IC (A); C absorbing', \
                      'Oblique Firehose', 'Parallel Firehose', \
                      'FM (B), oblique', 'FM (B), oblique; mirror']
    for cluster, ins_type in zip(np.arange(12), ins_types_cba_):
        df_cba_uns_post_SC_['ins_type'][df_cba_uns_post_SC_['cluster'] == cluster] = ins_type
    
    df_cba_post_SC_ = df_cba_.copy()
    df_cba_post_SC_['ins_type'] = np.nan
    df_cba_post_SC_.loc[df_cba_uns_post_SC_.index, ['ins_type']] = df_cba_uns_post_SC_['ins_type']
    
    return df_cba_post_SC_
    